from datetime import datetime
from threading import Thread, Lock
from tqdm import tqdm
import traceback
from pymongo import MongoClient

WAIT = 'wait'
RUNNING = 'running'
SUCCESS = 'success'
ERROR = 'error'


class MGSaver:
    def __init__(self, mongodb_connect: str, dbname='savedata'):
        self.getmg_func = lambda: MongoClient(mongodb_connect)
        self.dbname = dbname
        self._client = self.getmg_func()
        self._lock = Lock()

    # 保存任务输入参数
    def keys_save(self, keyname, keys: list, lies=None, if_replace_index=False):
        self._keyname = keyname
        self._kcet = self._client[self.dbname][keyname]
        if not lies:
            lies = keys[0].keys()
        if if_replace_index:
            try:
                self._kcet.drop_index(self._keyname)
            except:
                pass
        self._kcet.create_index(
            {lie: 1 for lie in lies}, unique=True, name=self._keyname)
        for dt in keys:
            try:
                dt['_complete'] = WAIT
                self._kcet.insert_one(dt)
            except:
                pass

    # 将结果表作为下个任务的参数表,包含全部数据
    def set_alldata_keytable(self, keyname):
        self._keyname = keyname
        self._kcet = self._client[self.dbname][keyname]
        self._kcet.update_many({'_complete': {'$ne': SUCCESS}}, {
                               '$set': {'_complete': WAIT}})

    # 获取匹配包含值的数据
    def getMongoContainDatadts(self, keyname, mark_lie:str, *values:str, lies:list = None, is_case_insensitive=True)->list:
        return self.getMongoDatadts(keyname, { '$or': [{mark_lie: {"$regex": value, '$options': 'i' if is_case_insensitive else None}}
                                                       for value in values]}, lies)
    
    def getMongoDatadts(self, keyname, query: dict, lies:list = None)->list:
        kcet = self._client[self.dbname][keyname]
        fields={'_id': 0}
        if lies: fields.update({lie: 1 for lie in lies})
        return list(kcet.find(query, fields))
        
    def _run(self):
        client = self.getmg_func()
        kcet = client[self.dbname][self._keyname]
        dcet = client[self.dbname][self._dataname]

        while True:
            try:
                # 取出数据
                with self._lock:
                    kdt = kcet.find_one_and_update({'_complete': WAIT},
                                                   {'$set': {'_complete': RUNNING, '_sd': datetime.now()}})
                if kdt is None: break
                taskid = kdt.pop('_id')
                # 移除框架自带的字段
                [kdt.pop(k) for k in list(kdt.keys()) if k.startswith('_')]
                dts = self._task_func(**kdt)
                if dts: dcet.insert_many([{'_key': taskid, **dt} for dt in dts])
                # 更新完成状态
                status = {'_complete': SUCCESS, '_ed': datetime.now()}
            except:
                err = traceback.format_exc()
                # 更新错误状态
                status = {'_complete': ERROR,
                          '_ed': datetime.now(), '_err': err}
            kcet.update_one({'_id': taskid}, {'$set': status})
            self._pbar.update()

        client.close()

    def task_run(self, dataname, kwargs_func_dts, threadnum=5, max_retry_num: int = 3):
        if max_retry_num <= 0: raise ValueError('重试次数已达上限!')
        self._dataname = dataname
        self._task_func = kwargs_func_dts
        task_num = self._kcet.count_documents({})

        # 初始化开始状态
        self._kcet.update_many({'_complete': {'$ne': SUCCESS, '$exists': True}}, {
                               '$set': {'_complete': WAIT}})
        self._pbar = tqdm(total=task_num, desc=f"task-{self._keyname}", unit="item")
        self._pbar.update(self._kcet.count_documents({'_complete': SUCCESS}))
        # 开始多线程任务
        ts = [Thread(target=self._run) for _ in range(threadnum)]
        [t.start() for t in ts]
        [t.join() for t in ts]
        self._pbar.close()
        # 监测任务完成情况
        incomplete_num = self._kcet.count_documents({'_complete': {'$ne': SUCCESS, '$exists': True}})
        if incomplete_num:
            print('未完成任务数量:', incomplete_num)
            self.task_run(dataname, kwargs_func_dts, threadnum=threadnum, max_retry_num=max_retry_num-1)
        else:
            print(f'任务-{self._keyname} 全部完成')

    def get_linkDataIt(self, keyname, dataname, tqdm_print_num=1000, is_keyname_startswith=True):
        with self.getmg_func() as client:
            dcet = client[self.dbname][dataname]
            desc = f'获取连接数据'
            data_num = dcet.count_documents({})
            if data_num == 0: 
                print('数据表为空!')
                return
            pbar = tqdm(total=data_num, desc=desc)
            dts = []
            for dt in dcet.aggregate([{'$lookup': {'from': keyname, 'localField': '_key', 'foreignField': '_id', 'as': '_keys'}}]):
                [dt.pop(k) for k in list(dt.keys())
                if k.startswith('_') and k != '_keys']
                keydt = dt.pop('_keys')[0]
                for k in list(keydt.keys()):
                    if k.startswith('_'):
                        keydt.pop(k)
                    else:
                        keydt[f'{keyname}_{k}' if is_keyname_startswith else k] = keydt.pop(k)
                dts.append({**keydt, **dt})
                if len(dts) == tqdm_print_num:
                    yield dts
                    pbar.update(len(dts))
                    dts.clear()
            if dts:
                yield dts
                pbar.update(len(dts))
            pbar.close()
    
    # 保存合并连接数据
    def save_linkData(self, keyname, dataname, linkname:str, tqdm_print_num=1000, is_keyname_startswith=True):
        client = self.getmg_func()
        lcet = client[self.dbname][linkname]
        lcet.drop()
        print(f'重新生成连接表-{linkname}')
        for dts in self.get_linkDataIt(keyname, dataname, tqdm_print_num=tqdm_print_num, is_keyname_startswith=is_keyname_startswith):
            lcet.insert_many(dts)

        client.close()


import pika
import logging
import logging.handlers
from threading import Thread
import json
import os
import time
import asyncio
import traceback

EXCHANGE = 'mengling_factory'

# 便捷获取log对象
def getLogger(name, level=logging.INFO, log_path=None) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.handlers.RotatingFileHandler(log_path,
                                                   maxBytes=5 * 1024 * 1024,
                                                   backupCount=5) if log_path else logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger

def getMQConnection(host, username, password, port=5672, virtual_host='/') -> pika.ConnectionParameters:
    # RabbitMQ服务器的连接参数
    parameters = pika.ConnectionParameters(
        host=host,  # RabbitMQ服务器地址
        port=port,         # RabbitMQ服务器端口，默认是5672
        virtual_host=virtual_host,  # 虚拟主机名称，默认是'/'
        credentials=pika.PlainCredentials(
            username=username,  # RabbitMQ用户名
            password=password  # RabbitMQ密码
        )
    )
    return parameters

def _init_factory(channel, factory, durable):
    channel.exchange_declare(exchange=EXCHANGE, exchange_type='topic')
    channel.queue_declare(queue=factory)
    channel.queue_bind(exchange=EXCHANGE, durable=durable, queue=factory, routing_key=f'{factory}_.*')
    
def _init_result(channel, factory, task_name, durable):
    queue_name = f'result_{factory}_{task_name}'
    channel.queue_declare(queue=queue_name)
    channel.queue_bind(exchange=EXCHANGE, durable=durable, queue=queue_name)
    return queue_name
            
def factory_start(mqcon: pika.ConnectionParameters, factory:str, good_getResult, 
                  good_num=1, logger: logging.Logger = None, durable=True)->list:
    logger = logger if logger else getLogger(f'factory-{factory}')
    logger.info(f'factory-{factory} worker num: {good_num}')
    
    def callback(ch, method, properties, body):
        try:
            result = good_getResult(body)
            ch.basic_publish(exchange=EXCHANGE, routing_key=f'result_{method.routing_key}', body=result)
        except:
            pass
    
    with pika.BlockingConnection(mqcon) as connection:
        channel = connection.channel()    
        _init_factory(channel, factory, durable=durable)
        def _worker():
            channel.basic_consume(queue=factory, on_message_callback=callback)
            channel.start_consuming()
        
        ts = [Thread(target=_worker, daemon=True) for _ in range(good_num)]
        [t.start() for t in ts]
        return ts


def task_puts(mqcon: pika.ConnectionParameters, factory:str, task_name:str, values:list, durable=True):
    with pika.BlockingConnection(mqcon) as connection:
        with connection.channel() as channel:
            _init_factory(channel, factory, durable=durable)
            queue_name = _init_result(channel, factory, task_name, durable=durable)
            for v in values:
                channel.basic_publish(exchange=EXCHANGE, routing_key=queue_name, body=v)

            
def result_It_get(mqcon: pika.ConnectionParameters, factory:str, task_name:str, durable=True)-> bytes:
    queue_name = _init_result(channel, factory, task_name, durable=durable)
    with pika.BlockingConnection(mqcon) as connection:
        with connection.channel() as channel:
            method, properties, body = channel.basic_get(queue=queue_name)
            if method:
                channel.basic_ack(delivery_tag = method.delivery_tag)
                yield body
            else:
                time.sleep(1)

async def async_result_It_get(mqcon: pika.ConnectionParameters, factory:str, task_name:str, durable=True)-> bytes:
    queue_name = _init_result(channel, factory, task_name, durable=durable)
    with pika.BlockingConnection(mqcon) as connection:
        with connection.channel() as channel:
            while True:
                method, properties, body = channel.basic_get(queue=queue_name)
                if method:
                    channel.basic_ack(delivery_tag = method.delivery_tag)
                    yield body
                # elif channel.(queue=f'{factory}_{task_name}') > 0:
                #     await asyncio.sleep(1)
                else:
                    break

def result_gets(mqcon: pika.ConnectionParameters):
    with pika.BlockingConnection(mqcon) as connection:
        with connection.channel() as channel:
            pass


if __name__ == '__main__':
    mqcon = getMQConnection('tianyiblue.work', 'ljh', 'mq246822', port=4672)
    with pika.BlockingConnection(mqcon) as connection:
        with connection.channel() as channel:
            queue_name = 'test'
            # channel.exchange_declare(exchange=EXCHANGE)
            # channel.queue_declare(queue=queue_name)
            # channel.queue_bind(exchange=EXCHANGE, queue=queue_name)
            # channel.basic_publish(exchange=EXCHANGE, routing_key=queue_name, body=json.dumps({'a': 3, 'b': 2}))
            method, properties, body = channel.basic_get(queue=queue_name)
            print(method, properties, body)
            channel.basic_nack(method.delivery_tag,requeue=True)
            # channel.basic_get(queue=queue_name)
            # print(result)



from threading import Thread, get_ident
from menglingtool.queue import Mqueue
import traceback
import os
import logging
import time
import asyncio

# 便捷获取log对象
def getLogger(name, level=logging.INFO, log_path=None) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.FileHandler(log_path) if log_path else logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


class WorkPool:
    def __init__(self, logger: logging.Logger = None) -> None:
        self.logger = logger if logger else getLogger(f'pid-{get_ident()}')
        self.task_queue = Mqueue()
        self._alive = False
    
    def getTaskNum(self) -> int:
        return self.task_queue.qsize()
    
    def is_alive(self) -> bool:
        return self._alive
    
    def stop(self):
        self._alive = False
    
    def start(self, *, getGood = None, getResult, good_num: int = 3,
                        is_put_args=False,
                        is_put_kwargs=False,
                        is_put_time=True) -> list:
        assert not self._alive, '任务已启动!'
        self.logger.info(f'PID-{os.getpid()} worker num: {good_num}')
        self._alive = True
        
        def _worker():
            gooder = getGood() if getGood else None
            while self._alive:
                que, index, args, kwargs = self.task_queue.get()
                sd = time.time()
                try:
                    result = getResult(gooder, *args, **kwargs) if gooder else getResult(*args, **kwargs)
                    err = None
                except:
                    err = traceback.format_exc()
                    result = None

                self.logger.info(f'args:{args}, kwargs:{kwargs}, is_err:{bool(err)}')
                rdt = {}
                if is_put_args: rdt['args'] = args
                if is_put_kwargs: rdt['kwargs'] = kwargs
                rdt['index'] = index
                rdt['result'] = result
                rdt['err'] = err
                if is_put_time: rdt['time'] = time.time() - sd
                que.put(rdt)

        ts = [Thread(target=_worker, daemon=True) for _ in range(good_num)]
        [t.start() for t in ts]
        return ts


    def arg_in_task_puts(self, vs: list) -> Mqueue: 
        return self.all_in_task_puts([[(v,), {}] for v in vs])


    def args_in_task_puts(self, argss: list) -> Mqueue:
        return self.all_in_task_puts([[args, {}] for args in argss])


    def kwargs_in_task_puts(self, kwargss: list) -> Mqueue:
        return self.all_in_task_puts([[(), kwargs] for kwargs in kwargss])


    def all_in_task_puts(self, args_and_kwargs: list) -> Mqueue:
        result_queue = Mqueue(maxsize=len(args_and_kwargs))
        self.task_queue.puts(*[(result_queue, i, *args_kwargs) for i, args_kwargs in enumerate(args_and_kwargs)])
        return result_queue

    @staticmethod
    def getResults(result_queue: Mqueue, is_sorded=True) -> list:
        while not result_queue.full():
            time.sleep(1)
        ls = result_queue.to_list()
        return sorted(ls, key = lambda x: x['index']) if is_sorded else ls

    @staticmethod
    async def async_getResults(result_queue: Mqueue, is_sorded=True) -> list:
        while not result_queue.full():
            await asyncio.sleep(1)
        ls = result_queue.to_list()
        return sorted(ls, key = lambda x: x['index']) if is_sorded else ls


from menglingtool_redis.redis_tool import RedisExecutor
import json
import base64
from menglingtool.thread import thread_auto_run


def getEncode(data) -> str:
    js = json.dumps(data, ensure_ascii=False)
    return base64.b64encode(js.encode('utf-8')).decode('utf-8')


def getDecode(enc):
    decoded_data = base64.b64decode(enc).decode('utf-8')
    return json.loads(decoded_data)


class Node:
    def __init__(self, name, last_node, dt_func_ls,
                 thread_num=3):
        self.name = name
        self.last_node = last_node
        self.func = dt_func_ls
        self.getR = last_node.getR
        self.thread_num = thread_num

    def getInEncs(self):
        return self.last_node.getOutEncs()

    def getOutEncs(self) -> set:
        rst = set()
        with self.getR() as r:
            for enc in r.hvals(self.name):
                rst.update(enc.split('\n'))
        return rst

    def _ceil(self, enc):
        datas = self.func(getDecode(enc))
        value = '\n'.join(getEncode(data) for data in datas)
        with self.getR() as r:
            r.hset(self.name, enc, value)

    def run(self, restart=False):
        # 获取上节点的全部值
        encs_all = self.getInEncs()
        print(f'节点-{self.name} 输入数据量:{len(encs_all)}')
        with self.getR() as r:
            if restart:
                print('删除节点记录数据,重新开始任务...')
                r.delete(self.name)
            else:
                encs_have = set(r.hkeys(self.name))
                print(f'已完成数据量:{len(encs_have)}')
        thread_auto_run(self._ceil, encs_all - encs_have,
                        threadnum=self.thread_num, max_error_num=3)


# 首节点
class FNode(Node):
    def __init__(self, name, datas, dt_func_ls, r_connect,
                 thread_num=3):
        self.getR = lambda: RedisExecutor(**r_connect)
        super().__init__(name, self, dt_func_ls, thread_num=thread_num)
        self.datas = datas

    def getInEncs(self):
        return set(getEncode(data) for data in self.datas)


# 尾结点
class LNode(Node):
    def __init__(self, last_node):
        super().__init__(Node, last_node, None)

    def resultsit(self, num=10_000) -> list:
        encs = list(self.getInEncs())
        for i in range(0, len(encs), num):
            yield [getDecode(enc) for enc in encs[i: i + num]]

    def getOutEncs(self):
        return []

    def run(self):
        pass
